# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

# Reuse the Linux code
from ._linux import LinuxJVMFinder

# ------------------------------------------------------------------------------


class DarwinJVMFinder(LinuxJVMFinder):
    """
    Mac OS X JVM library finder class
    """

    def __init__(self):
        """
        Sets up members
        """
        # Call the parent constructor
        LinuxJVMFinder.__init__(self)

        # Library file name
        self._libfile = "libjli.dylib"

        self._methods = list(self._methods)
        self._methods.append(self._pre_vm7_path)
        self._methods.append(self._javahome_binary)

        # Predefined locations
        self._locations = ('/Library/Java/JavaVirtualMachines',)

    def _pre_vm7_path(self):
        """
        Returns the previous constant JVM library path:
        '/System/Library/Frameworks/JavaVM.framework/JavaVM'
        """
        return '/System/Library/Frameworks/JavaVM.framework/JavaVM'

    def _javahome_binary(self):
        """
        for osx > 10.5 we have the nice util /usr/libexec/java_home available. Invoke it and
        return its output. It seems this tool has been removed in osx 10.9.
        """
        import platform
        import subprocess
        from distutils.version import StrictVersion

        current = StrictVersion(platform.mac_ver()[0][:4])
        if current >= StrictVersion('10.6') and current < StrictVersion('10.9'):
            if hasattr(subprocess, 'check_output'):
                java_home = subprocess.check_output(
                    ['/usr/libexec/java_home']).strip()
            else:
                java_home = subprocess.Popen(
                    ['/usr/libexec/java_home'], stdout=subprocess.PIPE).communicate()[0]
            return java_home

# ------------------------------------------------------------------------------


# Alias
JVMFinder = DarwinJVMFinder



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

from . import _jclass
from . import _jcustomizer
from . import _jtypes

JOverride = _jclass.JOverride


def isPythonSequence(v):
    if isinstance(v, Sequence):
        if not hasattr(v.__class__, '__metaclass__') \
           or v.__class__.__metaclass__ is _jclass._JavaClass:
            return True
    return False


@_jcustomizer.JImplementationFor("java.lang.Iterable")
class _JIterable(object):
    """ Customizer for ``java.util.Iterable``

    This customizer adds the Python iterator syntax to classes that 
    implement Java Iterable.
    """

    def __iter__(self):
        return self.iterator()


@_jcustomizer.JImplementationFor("java.util.Collection")
class _JCollection(object):
    """ Customizer for ``java.util.Collection``

    This customizer adds the Python functions ``len()`` and ``del`` to
    Java Collions to allow for Python syntax.
    """

    def __len__(self):
        return self.size()

    def __delitem__(self, i):
        return self.remove(i)

    @JOverride(sticky=True)
    def addAll(self, v):
        if isPythonSequence(v):
            v = _jclass.JClass('java.util.Arrays').asList(v)
        return self._addAll(v)

    @JOverride(sticky=True)
    def removeAll(self, v):
        if isPythonSequence(v):
            v = _jclass.JClass('java.util.Arrays').asList(v)
        return self._removeAll(v)

    @JOverride(sticky=True)
    def retainAll(self, v):
        if isPythonSequence(v):
            v = _jclass.JClass('java.util.Arrays').asList(v)
        return self._retainAll(v)


@_jcustomizer.JImplementationFor('java.util.List')
class _JList(object):
    """ Customizer for ``java.util.List``

    This customizer adds the Python list operator to function on classes
    that implement the Java List interface.
    """

    def __getitem__(self, ndx):
        if isinstance(ndx, slice):
            start = ndx.start
            stop = ndx.stop
            if start < 0:
                start = self.size() + start
            if stop < 0:
                stop = self.size() + stop
            return self.subList(start, stop)
        else:
            if ndx < 0:
                ndx = self.size() + ndx
            return self.get(ndx)

    def __setitem__(self, ndx, v):
        if isinstance(ndx, slice):
            if isPythonSequence(v):
                v = _jclass.JClass('java.util.Arrays').asList(v)
            self[ndx.start:ndx.stop].clear()
            self.addAll(start, v)
        else:
            if ndx < 0:
                ndx = self.size() + ndx
            self.set(ndx, v)

    def __delitem__(self, ndx):
        if isinstance(ndx, slice):
            self[ndx.start:ndx.stop].clear()
        elif hasattr(ndx, '__index__'):
            return self.remove(_jtypes.JInt(ndx))
        else:
            raise TypeError("Incorrect arguments to del")

    @JOverride(sticky=True)
    def addAll(self, *args):
        if len(args) == 1:
            v = args[0]
            if isPythonSequence(v):
                v = _jclass.JClass('java.util.Arrays').asList(v)
            self._addAll(v)
        elif len(args) == 2 and hasattr(args[0], '__index__'):
            v = args[1]
            if isPythonSequence(v):
                v = _jclass.JClass('java.util.Arrays').asList(v)
            self._addAll(args[0], v)
        else:
            raise TypeError("Incorrect arguments to addAll")


def isPythonMapping(v):
    if isinstance(v, collections.Mapping):
        if not hasattr(v.__class__, '__metaclass__') or \
           v.__class__.__metaclass__ is _jclass._JavaClass:
            return True
    return False


@_jcustomizer.JImplementationFor('java.util.Map')
class _JMap(object):
    """ Customizer for ``java.util.Map``

    This customizer adds the Python list and len operators to classes
    that implement the Java Map interface.
    """
    #    def __jclass_init__(cls):
    #        type.__setattr__(cls, 'putAll', _JMap.putAll)

    def __len__(self):
        return self.size()

    def __iter__(self):
        return self.keySet().iterator()

    def __delitem__(self, i):
        return self.remove(i)

    def __getitem__(self, ndx):
        return self.get(ndx)

    def __setitem__(self, ndx, v):
        self.put(ndx, v)

    @JOverride(sticky=True)
    def putAll(self, v):
        if isPythonMapping(v):
            for i in v:
                self.put(i, v[i])
        else:
            # do the regular method ...
            self._putAll(v)


@_jcustomizer.JImplementationFor('java.util.Iterator')
class _JIterator(object):
    """ Customizer for ``java.util.Iterator``

    This customizer adds the Python iterator concept to classes
    that implement the Java Iterator interface.
    """
#    def __jclass_init__(cls):
    #        type.__setattr__(cls, '_next', cls.next)
    #        type.__setattr__(cls, 'next', _JIterator.__next__)

    # Python 2 requires next to function as python next(), thus
    # we have a conflict in behavior. Java next is renamed.
    @JOverride(sticky=True, rename="_next")
    def next(self):
        if self.hasNext():
            return self._next()
        raise StopIteration

    def __next__(self):
        if self.hasNext():
            return self._next()
        raise StopIteration

    def __iter__(self):
        return self


@_jcustomizer.JImplementationFor('java.util.Enumeration')
class _JEnumeration(object):
    """ Customizer for ``java.util.Enumerator``

    This customizer adds the Python iterator concept to classes
    that implement the Java Enumerator interface.
    """

    def __next__(self):
        if self.hasMoreElements():
            return self.nextElement()
        raise StopIteration

    def __iter__(self):
        return self

    next = __next__



# -*- coding: utf-8 -*-
# *****************************************************************************
#   Copyright 2017 Karl Einar Nelson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

"""
JPype Imports Module
--------------------

Once imported this module will place the standard TLDs into the python
scope. These tlds are ``java``, ``com``, ``org``, and ``gov``.  Java
symbols from these domains can be imported using the standard Python
syntax.

Import customizers are supported in Python 3.6 or greater.

Forms supported:
   - **import <java_pkg> [ as <name> ]**
   - **import <java_pkg>.<java_class> [ as <name> ]**
   - **from <java_pkg> import <java_class>[,<java_class>*]**
   - **from <java_pkg> import <java_class> [ as <name> ]**
   - **from <java_pkg>.<java_class> import <java_static> [ as <name> ]**
   - **from <java_pkg>.<java_class> import <java_inner> [ as <name> ]**

For further information please read the :doc:`imports` guide.

Requires:
    Python 2.7 or 3.6 or later

Example:

.. code-block:: python

   import jpype
   import jpype.imports
   jpype.startJVM()

   # Import java packages as modules
   from java.lang import String

"""

import _jpype
try:
    from importlib.machinery import ModuleSpec as _ModuleSpec
    from types import ModuleType as _ModuleType
except Exception:

    # For Python2 compatiblity
    #  (Note: customizers are not supported)
    class _ModuleSpec(object):
        def __init__(self, name, loader):
            self.name = name
            self.loader = loader
    _ModuleType = object

import _jpype
import sys as _sys
from . import _pykeywords
from . import _jclass
from . import _jinit

__all__ = ["registerImportCustomizer", "registerDomain", "JImportCustomizer"]
_exportTypes = ()
_modifier = None

# %% Utility


def _keywordUnwrap(name):
    if not name.endswith('_'):
        return name
    if name[:-1] in _pykeywords._KEYWORDS:
        return name[:-1]
    return name


def _keywordWrap(name):
    if name in _pykeywords._KEYWORDS:
        return name + "_"
    return name


_java_lang_Class = None
_java_lang_NoClassDefFoundError = None
_java_lang_ClassNotFoundException = None
_java_lang_UnsupportedClassVersionError = None
def _getJavaClass(javaname):
    global _java_lang_Class
    global _java_lang_NoClassDefFoundError
    global _java_lang_ClassNotFoundException
    global _java_lang_UnsupportedClassVersionError
    if not _java_lang_Class:
      _java_lang_Class = _jclass.JClass("java.lang.Class")
      _java_lang_ClassNotFoundException = _jclass.JClass("java.lang.ClassNotFoundException")
      _java_lang_NoClassDefFoundError = _jclass.JClass("java.lang.NoClassDefFoundError")
      _java_lang_UnsupportedClassVersionError = _jclass.JClass("java.lang.UnsupportedClassVersionError")

    err = None
    try:
        # Use forname because it give better diagnostics
        cls = _java_lang_Class.forName(javaname)
        return _jclass.JClass(cls)

    # Not found is acceptable
    except _java_lang_ClassNotFoundException:
        return None

    # Missing dependency
    except _java_lang_NoClassDefFoundError as ex:
        missing = str(ex).replace('/','.')
        err = "Unable to import '%s' due to missing dependency '%s'"%(javaname, missing)

    # Wrong Java version
    except _java_lang_UnsupportedClassVersionError as ex:
        err = "Unable to import '%s' due to incorrect Java version"%(javaname)

    # Otherwise!?
    except Exception as ex:
        err = "Unable to import '%s' due to unexpected exception, '%s'"%(javaname, ex)
    raise ImportError(err)

# FIXME imports of static fields not working for now.


def _copyProperties(out, mc):
    #    for jf in mc.__javaclass__.getClassFields():
    #        out[_keywordWrap(jf.getName())] = jf
    pass


def _getStaticMethods(cls):
    global _modifier
    static = {}
    for u in cls.class_.getMethods():
        if not _modifier.isStatic(u.getModifiers()):
            continue
        name = _keywordWrap(str(u.getName()))
        static[name] = getattr(cls, name)
    return static


def _copyStaticMethods(out, cls):
    for u, v in _getStaticMethods(cls).items():
        out[u] = v


# %% Customizer
_CUSTOMIZERS = []

if _sys.version_info > (3,):
    def registerImportCustomizer(customizer):
        """ Import customizers can be used to import python packages
        into java modules automatically.
        """
        _CUSTOMIZERS.append(customizer)

    # Support hook for placing other things into the java tree
    class JImportCustomizer(object):
        """ Base class for Import customizer.

        Import customizers should implement canCustomize and getSpec.

        Example:

        .. code-block:: python

           # Site packages for each java package are stored under $DEVEL/<java_pkg>/py
           class SiteCustomizer(jpype.imports.JImportCustomizer):
               def canCustomize(self, name):
                   if name.startswith('org.mysite') and name.endswith('.py'):
                       return True
                   return False
               def getSpec(self, name):
                   pname = name[:-3]
                   devel = os.environ.get('DEVEL')
                   path = os.path.join(devel, pname,'py','__init__.py')
                   return importlib.util.spec_from_file_location(name, path)
       """

        def canCustomize(self, name):
            """ Determine if this path is to be treated differently 

            Return:
                True if an alternative spec is required.
            """
            return False

        def getSpec(self, name):
            """ Get the module spec for this module. 
            """
            raise NotImplementedError
else:
    def registerImportCustomizer(customizer):
        raise NotImplementedError(
            "Import customizers not implemented for Python 2.x")
    JImportCustomizer = object


# %% Import
class _JImport(object):
    """ (internal) Base class for import java modules """
    # Module requirements
    __doc__ = None
    __loader__ = None
    __path__ = []
    __package__ = "java"

    def __init__(self, name):
        pass

    def __getattr__(self, name):
        if name.startswith('_'):
            return object.__getattribute__(self, name)

        name = _keywordUnwrap(name)

        # Inner class support
        jname = object.__getattribute__(self, '__javaname__')
        try:
            object.__getattribute__(self, '__javaclass__')
            jname = "$".join([jname, name])
        except AttributeError:
            jname = ".".join([jname, name])

        # Get the class (if it exists)
        jtype = _getJavaClass(jname)
        if jtype:
            # Cache it for later
            object.__setattr__(self, name, jtype)
            return jtype

        # If the java class does not exist, throw a ClassNotFound exception
        raise ImportError("Unable to find java class '%s'"%jname)

    def __setattr__(self, name, value):
        if name.startswith('__'):
            raise AttributeError("Module does not allow setting of '%s'" % name)
        if hasattr(value, '__javaclass__'):
            return object.__setattr__(self, name, getattr(value, '__javaclass__'))
        if isinstance(value, (_JImport, _ModuleType)):
            return object.__setattr__(self, name, value)
        raise AttributeError("JImport may not set attribute '%s'" % name)


# In order to get properties to be attached to the _JImport class,
# we must create a dynamic class between
def _JImportFactory(spec, javaname, cls=_JImport):
    """ (internal) Factory for creating java modules dynamically.

    This is needed to create a new type node to hold static methods.
    """

    def init(self, name):
        # Call the base class
        cls.__init__(self, name)

    def getall(self):
        global _exportTypes
        d1 = self.__dict__.items()
        d2 = self.__class__.__dict__.items()
        local = [name for name, attr in d1 if not name.startswith('_')
                 and isinstance(attr, _exportTypes)]
        glob = [name for name, attr in d2 if not name.startswith('_')
                and isinstance(attr, _exportTypes)]
        local.extend(glob)
        return local

    # Set up a new class for this type
    bases = [cls]
    members = {
        "__init__": init,
        "__javaname__": javaname,
        "__name__": spec.name,
        "__all__": property(getall),
        "__spec__": spec,
    }

    # Is this module also a class, if so insert class info
    jclass = _getJavaClass(javaname)
    if jclass:
        # Mark this as a class (will cause children to be inner classes)
        members['__javaclass__'] = jclass

        # Exposed static members as part of the module
        _copyProperties(members, jclass)
        _copyStaticMethods(members, jclass)

    return type("module." + spec.name, tuple(bases), members)


def _JModule(spec, javaname):
    """ (internal) Front end for creating a java module dynamically """
    cls = _JImportFactory(spec, javaname)
    out = cls(spec.name)
    return out

# %% Finder


class _JImportLoader:
    """ (internal) Finder hook for importlib. """

    def find_spec(self, name, path, target):
        parts = name.split('.', 1)
        if not parts[0] in _JDOMAINS:
            return None

        # Support for external modules in java tree
        for customizer in _CUSTOMIZERS:
            if customizer.canCustomize(name):
                return customizer.getSpec(name)

        # Import the java module
        return _ModuleSpec(name, self)

    """ (internal) Loader hook for importlib. """

    def create_module(self, spec):
        if not _jpype.isStarted():
            raise ImportError("Attempt to create java modules without jvm")

        # Handle creating the java name based on the path
        parts = spec.name.split('.')
        if len(parts) == 1:
            return _JModule(spec, _JDOMAINS[spec.name])

        # Use the parent module to simplify name mangling
        base = _sys.modules[".".join(parts[:-1])]

        # Support of inner classes
        if not isinstance(base, _JImport):
            return getattr(base, parts[-1])
        jbasename = object.__getattribute__(base, '__javaname__')
        try:
            object.__getattribute(base, '__javaclass__')
            javaname = "$".join([jbasename, _keywordUnwrap(parts[-1])])
        except AttributeError:
            javaname = ".".join([jbasename, _keywordUnwrap(parts[-1])])

        return _JModule(spec, javaname)

    def exec_module(self, fullname):
        pass

    # For compatablity with Python 2.7
    def find_module(self, name, path=None):
        parts = name.split('.', 1)
        if not parts[0] in _JDOMAINS:
            return None
        return self

    # For compatablity with Python 2.7
    def load_module(self, name):
        module = self.create_module(_ModuleSpec(name, self))
        _sys.modules[name] = module
        return module


# Install hooks into python importlib
_sys.meta_path.append(_JImportLoader())

# %% Domains
_JDOMAINS = {}


def registerDomain(mod, alias=None):
    """ Add a java domain to python as a dynamic module.

    This can be used to bind a Java path to a Python path.

    Args:
        mod(str): Is the Python module to bind to Java.
        alias(str, optional): Is the name of the Java path if different 
          than the Python name. 
    """
    if not alias:
        alias = mod
    _JDOMAINS[mod] = alias


# Preregister common top level domains
registerDomain('com')
registerDomain('gov')
registerDomain('java')
registerDomain('org')

# %% Initialize


def _initialize():
    global _exportTypes
    global _modifier
    _JMethod = type(_jclass.JClass('java.lang.Class').forName)
    _modifier = _jclass.JClass('java.lang.reflect.Modifier')
    _exportTypes = (property, _jclass.JClass, _JImport, _JMethod)


_jinit.registerJVMInitializer(_initialize)



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

from . import _jvmfinder

# ------------------------------------------------------------------------------


def _checkJVMArch(jvmPath):
    IMAGE_FILE_MACHINE_I386 = 332
    IMAGE_FILE_MACHINE_IA64 = 512
    IMAGE_FILE_MACHINE_AMD64 = 34404

    is64 = _sys.maxsize > 2**32
    with open(jvmPath, "rb") as f:
        s = f.read(2)
        if s != b"MZ":
            raise JVMNotSupportedException("JVM not valid")
        else:
            f.seek(60)
            s = f.read(4)
            header_offset = _struct.unpack("<L", s)[0]
            f.seek(header_offset+4)
            s = f.read(2)
            machine = _struct.unpack("<H", s)[0]

    if machine == IMAGE_FILE_MACHINE_I386:
        if is64:
            raise JVMNotSupportedException(
                "JVM mismatch, python is 64 bit and JVM is 32 bit.")
    elif machine == IMAGE_FILE_MACHINE_IA64 or machine == IMAGE_FILE_MACHINE_AMD64:
        if not is64:
            raise JVMNotSupportedException(
                "JVM mismatch, python is 32 bit and JVM is 64 bit.")
    else:
        raise JVMNotSupportedException("Unable to deterime JVM Type")


class WindowsJVMFinder(_jvmfinder.JVMFinder):
    """
    Windows JVM library finder class
    """

    def __init__(self):
        """
        Sets up members
        """
        # Call the parent constructor
        _jvmfinder.JVMFinder.__init__(self)

        # Library file name
        self._libfile = "jvm.dll"

        # Search methods
        self._methods = (self._get_from_java_home, self._get_from_registry)

    def check(self, jvm):
        from ._windows import _checkJVMArch
        _checkJVMArch(jvm)

    def _get_from_registry(self):
        """
        Retrieves the path to the default Java installation stored in the
        Windows registry

        :return: The path found in the registry, or None
        """
        from ._windows import reg_keys
        for location in reg_keys:
            location = location.replace('\\', '/')
            jreKey = "/proc/registry/HKEY_LOCAL_MACHINE/{}".format(location)
            try:
                with open(jreKey + "/CurrentVersion") as f:
                    cv = f.read().split('\x00')
                versionKey = jreKey + "/" + cv[0]

                with open(versionKey + "/RunTimeLib") as f:
                    cv = f.read().split('\x00')

                return cv[0]

            except OSError:
                pass
        return None



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

import collections
import sys

__all__ = ["JProxy", "JImplements"]

import _jpype
from . import _jclass

if sys.version > '3':
    unicode = str

# FIXME the java.lang.method we are overriding should be passes to the lookup function
# so we can properly handle name mangling on the override.


def _createJProxy(cls, *intf, **kwargs):
    """ (internal) Create a proxy from a python class with
    @JOverride notation on methods.
    """
    # Convert the interfaces list
    actualIntf = _convertInterfaces(intf)

    # Find all class defined overrides
    overrides = {}
    for k, v in cls.__dict__.items():
        try:
            attr = object.__getattribute__(v, "__joverride__")
            overrides[k] = (v, attr)
        except AttributeError:
            pass

    # Verify all methods are overriden
    for interface in actualIntf:
        for method in interface.class_.getMethods():
            if method.getModifiers() & 1024 == 0:
                continue
            if not str(method.getName()) in overrides:
                raise NotImplementedError("Interface '%s' requires method '%s' to be implemented." % (
                    interface.class_.getName(), method.getName()))

    # Define a lookup interface
    def lookup(self, name):
        # Get the override from the override dictionary
        over = overrides[name]

        # We need convert the method to a bound method using descriptor interface
        return over[0].__get__(self)

    # Construct a new init method
    init = cls.__dict__.get('__init__', None)
    if init:
        def init2(self, *args, **kwargs):
            init(self, *args, **kwargs)
            self.__javaproxy__ = _jpype.PyJPProxy(self, lookup, actualIntf)
    else:
        def init2(self, *args, **kwargs):
            self.__javaproxy__ = _jpype.PyJPProxy(self, lookup, actualIntf)

    # Replace the init with the proxy init
    type.__setattr__(cls, '__init__', init2)

    # Return the augmented class
    return cls


def JImplements(*args, **kwargs):
    """ Annotation for creating a new proxy that implements a list of
    Java interfaces.

    This annotation is placed on an ordinary Python class.  The annotation
    requires a list of interfaces.  It must implement all of the java
    methods for each of the interfaces.  Each implemented method
    should have a @JOverride annotation.

    Args:
      interfaces (str*,JClass*): Strings or JClasses for each Java interface
        this proxy is to implement.

    Example:

      .. code-block:: python

          @JImplement("java.lang.Runnable")
          class MyImpl(object):
             @JOverride
             def run(self, arg):
               pass

          @JImplement("org.my.Interface1", "org.my.Interface2")
          class MyImpl(object):
             @JOverride
             def method(self, arg):
               pass

    """
    def JProxyCreator(cls):
        return _createJProxy(cls, *args, **kwargs)
    return JProxyCreator


def _convertInterfaces(intf):
    """ (internal) Convert a list of interface names into
    a list of interfaces suitable for a proxy.
    """
    # Flatten the list
    intflist = []
    for item in intf:
        if isinstance(item, (str, unicode)) or not hasattr(item, '__iter__'):
            intflist.append(item)
        else:
            intflist.extend(item)

    # Look up the classes if given as a string
    actualIntf = set()
    for item in intflist:
        if isinstance(item, (str, unicode)):
            actualIntf.add(_jclass.JClass(item))
        else:
            actualIntf.add(item)

    # Check that all are interfaces
    if not actualIntf:
        raise TypeError("At least one Java interface must be specified")

    for cls in actualIntf:
        # If it isn't a JClass, then it cannot be a Java interface
        if not isinstance(cls, _jclass.JClass):
            raise TypeError("'%s' is not a Java interface"%type(cls).__name__)
        # Java concrete and abstract classes cannot be proxied
        if not issubclass(cls, _jclass.JInterface):
            raise TypeError("'%s' is not a Java interface"%cls.__name__)

    return tuple(actualIntf)


class JProxy(object):
    """ Define a proxy for a Java interface.

    This is an older style JPype proxy interface that uses either a
    dictionary or an object instance to implement methods defined
    in java.  The python object can be held by java and its lifespan
    will continue as long as java holds a reference to the object
    instance.  New code should use ``@JImplements`` annotation as
    it will support improved type safety and error handling.

    Name lookups can either made using a dictionary or an object
    instance.  One of these two options must be specified.

    Args:
        intf: either a single interface or a list of java interfaces.
            The interfaces can either be defined by strings or
            JClass instance.  Only interfaces may be used in a
            proxy,
        dict (dict[string, callable], optional): specifies a dictionary
            containing the methods to be called when executing the
            java interface methods.
        inst (object, optional): specifies an object with methods
            whose names matches the java interfaces methods.
    """

    def __init__(self, intf, dict=None, inst=None):
        # Convert the interfaces
        actualIntf = _convertInterfaces([intf])

        # Verify that one of the options has been selected
        if dict is not None and inst is not None:
            raise RuntimeError("Specify only one of dict and inst")

        if dict is not None:
            # Define the lookup function based for a dict
            def lookup(d, name):
                return d[name]
            # create a proxy
            self.__javaproxy__ = _jpype.PyJPProxy(dict, lookup, actualIntf)
            return

        if inst is not None:
            # Define the lookup function based for a object instance
            def lookup(d, name):
                return getattr(d, name)
            # create a proxy
            self.__javaproxy__ = _jpype.PyJPProxy(inst, lookup, actualIntf)
            return
        raise TypeError("a dict or inst must be specified")



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys

import _jpype
from . import _jclass
from . import _jobject
from . import _jcustomizer

__all__ = ['JBoolean', 'JByte', 'JChar', 'JShort',
           'JInt', 'JLong', 'JFloat', 'JDouble']

if _sys.version_info > (3,):
    _unicode = str
    _long = int
else:
    _unicode = unicode
    _long = long

# FIXME python2 and python3 get different conversions on int and long.  Likely we should
# unify to got the the same types regardless of version.

# Set up all the tables
_maxFloat = 0
_maxDouble = 0


def _initialize():
    _JP_TYPE_CLASSES = _jclass._JP_TYPE_CLASSES
    _JCLASSES = _jclass._JCLASSES
    global _maxFloat, _maxDouble

    # Place the jvalue types for primitives
    _JPrimitiveLoad(JBoolean, _jclass.JClass("java.lang.Boolean"))
    _JPrimitiveLoad(JByte, _jclass.JClass("java.lang.Byte"))
    _JPrimitiveLoad(JChar, _jclass.JClass("java.lang.Character"))
    _JPrimitiveLoad(JShort, _jclass.JClass("java.lang.Short"))
    _JPrimitiveLoad(JInt, _jclass.JClass("java.lang.Integer"))
    _JPrimitiveLoad(JLong, _jclass.JClass("java.lang.Long"))
    _JPrimitiveLoad(JFloat, _jclass.JClass("java.lang.Float"))
    _JPrimitiveLoad(JDouble, _jclass.JClass("java.lang.Double"))

    # Set up table of automatic conversions
    _JP_TYPE_CLASSES[bool] = JBoolean
    _JP_TYPE_CLASSES[int] = JLong
    _JP_TYPE_CLASSES[_long] = JLong
    _JP_TYPE_CLASSES[float] = JDouble
    _JP_TYPE_CLASSES[str] = _jclass.JClass("java.lang.String")
    _JP_TYPE_CLASSES[_unicode] = _jclass.JClass("java.lang.String")
    _JP_TYPE_CLASSES[type] = _jclass.JClass("java.lang.Class")
    _JP_TYPE_CLASSES[object] = _jclass.JClass("java.lang.Object")

    _maxFloat = _jclass.JClass("java.lang.Float").MAX_VALUE
    _maxDouble = _jclass.JClass("java.lang.Double").MAX_VALUE


def _JPrimitiveLoad(cls, boxedType):
    type.__setattr__(cls, '__javaclass__', _jpype.PyJPClass(cls.__name__))
    type.__setattr__(cls, '_java_boxed_class', boxedType)


class _JPrimitiveClass(_jclass.JClass):
    """ A wrapper specifying a specific java type.

    These objects have three fields:

     - __javaclass__ - the class for this object when matching arguments.
     - _java_boxed_class - the class to convert to when converting to an 
        object.
     - __javavalue__ - the instance of the java value.

    """
    def __new__(cls, name, basetype):
        members = {
            "__init__": _JPrimitive.init,
            "__setattr__": object.__setattr__,
            "__javaclass__": None,
            "_java_boxed_class": None,
        }
        return super(_JPrimitiveClass, cls).__new__(cls, name, (basetype, _JPrimitive), members)

    def __init__(self, *args):
        _jclass._JCLASSES[args[0]] = self
        super(_JPrimitive, self).__init__(self)

    def _load(self, boxed):
        type.__setattr__(self, '__javaclass__',
                         _jpype.PyJPClass(self.__name__))
        type.__setattr__(self, '_java_boxed_class', boxed)


class _JPrimitive(object):
    def __setattr__(self, attr, value):
        raise AttributeError("%s does not have field %s" %
                             (self.__name__, attr))

    def init(self, v):
        if v is not None:
            self._pyv = v
            self.__javavalue__ = _jpype.PyJPValue(self.__javaclass__, v)
        else:
            self.__javavalue__ = None

    def byteValue(self):
        if self._pyv < -128 or self._pyv > 127:
            raise OverFlowError("Cannot convert to byte value")
        return int(self._pyv)

    def shortValue(self):
        if self._pyv < -32768 or self._pyv > 32767:
            raise OverFlowError("Cannot convert to short value")
        return int(self._pyv)

    def intValue(self):
        if self._pyv < -2147483648 or self._pyv > 2147483647:
            raise OverFlowError("Cannot convert to int value")
        return int(self._pyv)

    def longValue(self):
        if self._pyv < -9223372036854775808 or self._pyv > 9223372036854775807:
            raise OverFlowError("Cannot convert to long value")
        return _long(self._pyv)

    def floatValue(self):
        if self._pyv < -_maxFloat or self._pyv > _maxFloat:
            raise OverFlowError("Cannot convert to long value")
        return float(self._pyv)

    def doubleValue(self):
        if self._pyv < -_maxDouble or self._pyv > _maxDouble:
            raise OverFlowError("Cannot convert to double value")
        return float(self._pyv)


JBoolean = _JPrimitiveClass("boolean", int)
JByte = _JPrimitiveClass("byte", int)
JChar = _JPrimitiveClass("char", int)
JShort = _JPrimitiveClass("short", int)
JInt = _JPrimitiveClass("int", int)
JLong = _JPrimitiveClass("long", _long)
JFloat = _JPrimitiveClass("float", float)
JDouble = _JPrimitiveClass("double", float)



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys

import _jpype

from . import _jclass
from . import _jobject
from . import _jtypes
from . import _classpath

# Import all the class customizers
# Customizers are applied in the order that they are defined currently.
#from . import _properties
from . import _jarray
from . import _jboxed
from . import _jexception
from . import _jcollection
from . import _jcomparable
from . import _jio
from . import _jinit
from ._jvmfinder import JVMNotFoundException, JVMNotSupportedException

__all__ = [
    'isJVMStarted', 'startJVM', 'attachToJVM', 'shutdownJVM',
    'getDefaultJVMPath', 'getJVMVersion', 'isThreadAttachedToJVM', 'attachThreadToJVM',
    'detachThreadFromJVM', 'synchronized', 'get_default_jvm_path',
    'JVMNotFoundException', 'JVMNotSupportedException'
]


# Activate jedi tab completion
try:
    import jedi as _jedi
    _jedi.evaluate.compiled.access.ALLOWED_DESCRIPTOR_ACCESS += \
       ( _jpype.PyJPMethod, _jpype.PyJPField)
except:
    pass


# See http://scottlobdell.me/2015/04/decorators-arguments-python/
def deprecated(*args):
    """ Marks a function a deprecated when used as decorator.

    Be sure to start python with -Wd to see warnings.
    """
    def func2(*args, **kwargs):
        import warnings
        if not func2._warned:
            warnings.warn(func2._warning % (func2._real.__module__, func2._real.__name__),
                          category=DeprecationWarning, stacklevel=2)
        func2._warned = True
        return func2._real(*args, **kwargs)

    if isinstance(args[0], str):
        def decorate(func):
            func2.__name__ = func.__name__
            func2.__doc__ = func.__doc__
            func2._warned = False
            func2._real = func
            func2._warning = "%s.%s is deprecated, use {0} instead".format(
                args[0])
            return func2
        return decorate
    else:
        func = args[0]
        func2.__name__ = func.__name__
        func2.__doc__ = func.__doc__
        func2._warned = False
        func2._real = func
        func2._warning = "%s.%s is deprecated"
        return func2


def _initialize():
    _jclass._initialize()
    _jobject._initialize()
    _jtypes._initialize()
    _jinit.runJVMInitializers()


def isJVMStarted():
    return _jpype.isStarted()


def _hasClassPath(args):
    for i in args:
        if i.startswith('-Djava.class.path'):
            return True
    return False

def _handleClassPath(clsList):
    out = []
    for s in clsList:
        if not isinstance(s, (str, _jtypes._unicode)):
            raise TypeError("Classpath elements must be strings")
        if s.endswith('*'):
            import glob
            out.extend(glob.glob(s+'.jar'))
        else:
            out.append(s)
    return _classpath._SEP.join(out)


_JVM_started = False

def startJVM(*args, **kwargs):
    """
    Starts a Java Virtual Machine.  Without options it will start
    the JVM with the default classpath and jvmpath.

    The default classpath is determined by ``jpype.getClassPath()``.
    The default jvmpath is determined by ``jpype.getDefaultJVMPath()``.

    Parameters:
     *args (Optional, str[]): Arguments to give to the JVM.
        The first argument may be the path the JVM.

    Keyword Arguments:
      jvmpath (str):  Path to the jvm library file,
        Typically one of (``libjvm.so``, ``jvm.dll``, ...)
        Using None will apply the default jvmpath.
      classpath (str,[str]): Set the classpath for the jvm.
        This will override any classpath supplied in the arguments
        list. A value of None will give no classpath to JVM.
      ignoreUnrecognized (bool): Option to JVM to ignore
        invalid JVM arguments. Default is False.
      convertStrings (bool): Option to JPype to force Java strings to
        cast to Python strings. This option is to support legacy code
        for which conversion of Python strings was the default. This
        will globally change the behavior of all calls using
        strings, and a value of True is NOT recommended for newly 
        developed code.

        The default value for this option during 0.7 series is 
        True.  The option will be False starting in 0.8. A
        warning will be issued if this option is not specified
        during the transition period.


    Raises:
      OSError: if the JVM cannot be started or is already running.
      TypeError: if an invalid keyword argument is supplied
        or a keyword argument conflicts with the arguments.

     """
    if _jpype.isStarted():
        raise OSError('JVM is already started')
    global _JVM_started
    if _JVM_started:
        raise OSError('JVM cannot be restarted')
    _JVM_started = True

    args = list(args)

    # JVM path
    jvmpath = None
    if args:
        # jvm is the first argument the first argument is a path or None
        if not args[0] or not args[0].startswith('-'):
            jvmpath = args.pop(0)
    if 'jvmpath' in kwargs:
        if jvmpath:
            raise TypeError('jvmpath specified twice')
        jvmpath = kwargs.pop('jvmpath')
    if not jvmpath:
        jvmpath = getDefaultJVMPath()

    # Classpath handling
    if _hasClassPath(args):
        # Old style, specified in the arguments
        if 'classpath' in kwargs:
            # Cannot apply both styles, conflict
            raise TypeError('classpath specified twice')
        classpath = None
    elif 'classpath' in kwargs:
        # New style, as a keywork
        classpath = kwargs.pop('classpath')
    else:
        # Not speficied at all, use the default classpath
        classpath = _classpath.getClassPath()

    # Handle strings and list of strings.
    if classpath:
        if isinstance(classpath, (str, _jtypes._unicode)):
            args.append('-Djava.class.path=%s' % _handleClassPath([classpath]))
        elif hasattr(classpath, '__iter__'):
            args.append('-Djava.class.path=%s' % _handleClassPath(classpath))
        else:
            raise TypeError("Unknown class path element")

    ignoreUnrecognized = kwargs.pop('ignoreUnrecognized', False)

    if not "convertStrings" in kwargs:
        import warnings
        warnings.warn("""
-------------------------------------------------------------------------------
Deprecated: convertStrings was not specified when starting the JVM. The default
behavior in JPype will be False starting in JPype 0.8. The recommended setting
for new code is convertStrings=False.  The legacy value of True was assumed for
this session. If you are a user of an application that reported this warning,
please file a ticket with the developer.
-------------------------------------------------------------------------------
""")

    convertStrings = kwargs.pop('convertStrings', True)


    if kwargs:
        raise TypeError("startJVM() got an unexpected keyword argument '%s'"
                        % (','.join([str(i) for i in kwargs])))

    _jpype.startup(jvmpath, tuple(args), ignoreUnrecognized, convertStrings)
    _initialize()


def attachToJVM(jvm):
    _jpype.attach(jvm)
    _initialize()


def shutdownJVM():
    """ Shuts down the JVM.

    This method shuts down the JVM and thus disables access to existing
    Java objects. Due to limitations in the JPype, it is not possible to
    restart the JVM after being terminated.
    """
    _jpype.shutdown()


def isThreadAttachedToJVM():
    """ Checks if a thread is attached to the JVM.

    Python automatically attaches threads when a Java method is called.
    This creates a resource in Java for the Python thread. This method
    can be used to check if a Python thread is currently attached so that
    it can be disconnected prior to thread termination to prevent leaks.

    Returns:
      True if the thread is attached to the JVM, False if the thread is
      not attached or the JVM is not running.
    """
    return _jpype.isThreadAttachedToJVM()


def attachThreadToJVM():
    """ Attaches a thread to the JVM.

    The function manually connects a thread to the JVM to allow access to
    Java objects and methods. JPype automaticatlly attaches when a Java
    resource is used, so a call to this is usually not needed.

    Raises:
      RuntimeError: If the JVM is not running.
    """
    _jpype.attachThreadToJVM()


def detachThreadFromJVM():
    """ Detaches a thread from the JVM.

    This function detaches the thread and frees the associated resource in
    the JVM. For codes making heavy use of threading this should be used
    to prevent resource leaks. The thread can be reattached, so there
    is no harm in detaching early or more than once. This method cannot fail
    and there is no harm in calling it when the JVM is not running.
    """
    _jpype.detachThreadFromJVM()


def synchronized(obj):
    """ Creates a resource lock for a Java object.

    Produces a monitor object. During the lifespan of the monitor the Java
    will not be able to acquire a thread lock on the object. This will
    prevent multiple threads from modifying a shared resource.

    This should always be used as part of a Python ``with`` startment.

    Arguments:
        obj: A valid Java object shared by multiple threads.

    Example:

    .. code-block:: python

      with synchronized(obj):
         # modify obj values

      # lock is freed when with block ends

    """
    try:
        return _jpype.PyJPMonitor(obj.__javavalue__)
    except AttributeError as ex:
        pass
    raise TypeError("synchronized only applies to java objects")


def getDefaultJVMPath():
    """
    Retrieves the path to the default or first found JVM library

    Returns:
      The path to the JVM shared library file

    Raises:
      JVMNotFoundException: If there was no JVM found in the search path.
      JVMNotSupportedException: If the JVM was found was not compatible with
        Python due to cpu architecture.

    """
    if _sys.platform == "cygwin":
        # Cygwin
        from ._cygwin import WindowsJVMFinder
        finder = WindowsJVMFinder()
    elif _sys.platform == "win32":
        # Windows
        from ._windows import WindowsJVMFinder
        finder = WindowsJVMFinder()
    elif _sys.platform == "darwin":
        # Mac OS X
        from ._darwin import DarwinJVMFinder
        finder = DarwinJVMFinder()
    else:
        # Use the Linux way for other systems
        from ._linux import LinuxJVMFinder
        finder = LinuxJVMFinder()

    return finder.get_jvm_path()


# Naming compatibility
@deprecated("getDefaultJVMPath")
def get_default_jvm_path(*args, **kwargs):
    return getDefaultJVMPath(*args, **kwargs)


def getJVMVersion():
    """ Get the JVM version if the JVM is started.

    This function can be used to determine the version of the JVM. It is
    useful to help determine why a Jar has failed to load.

    Returns:
      A typle with the (major, minor, revison) of the JVM if running.
    """
    if not _jpype.isStarted():
        return (0, 0, 0)

    import re
    runtime = _jclass.JClass('java.lang.Runtime')
    version = runtime.class_.getPackage().getImplementationVersion()

    # Java 9+ has a version method 
    if not version:
        version = runtime.version()
    version = (re.match("([0-9.]+)", str(version)).group(1))
    return tuple([int(i) for i in version.split('.')])




# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

import sys as _sys
import inspect

import _jpype
from . import _jclass
from . import _jcustomizer
from . import _jinit

__all__ = ['JObject']

if _sys.version > '3':
    _unicode = str
    _long = int
else:
    _unicode = unicode
    _long = long


def _initialize():
    type.__setattr__(JObject, '__javaclass__',
                     _jpype.PyJPClass('java.lang.Object'))


class JObject(object):
    """ Base class for all object instances.

    It can be used to test if an object is a java object instance with
    ``isinstance(obj, JObject)``.

    Calling ``JObject`` as a function can be used to covert or cast to 
    specific Java type.  It will box primitive types and supports an 
    option type to box to.

    This wrapper functions four ways. 

      - If the no type is given the object is automatically 
        cast to type best matched given the value.  This can be used 
        to create a boxed primitive.  ``JObject(JInt(i))``

      - If the type is a primitve, the object will be the boxed type of that
        primitive.  ``JObject(1, JInt)``

      - If the type is a Java class and the value is a Java object, the
        object will be cast to the Java class and will be an exact match to 
        the class for the purposes of matching arguments. If the object 
        is not compatible, an exception will be raised.

      - If the value is a python wrapper for class it will create a class
        instance.  This is aliased to be much more obvious as the ``class_`` 
        member of each Java class.

    Args:
       value: The value to be cast into an Java object.
       type(Optional, type): The type to cast into.  

    Raises:
       TypeError: If the object cannot be cast to the specified type, or
         the requested type is not a Java class or primitive.

    """
    def __new__(cls, *args, **kwargs):
        if cls != JObject:
            return super(JObject, cls).__new__(cls)
        # Create a null pointer object
        if len(args) == 0:
            args = [None]
        cls = _JObjectFactory(*args, **kwargs)
        self = cls.__new__(cls, args[0])
        self.__javavalue__ = _jpype.PyJPValue(cls.__javaclass__, args[0])
        return self

    def __init__(self, *args):
        if hasattr(self, '__javavalue__'):
            pass
        elif len(args) == 1 and isinstance(args[0], _jpype.PyJPValue):
            object.__setattr__(self, '__javavalue__', args[0])
        elif not hasattr(self, '__javavalue__'):
            jv = self.__class__.__javaclass__.newInstance(*args)
            object.__setattr__(self, '__javavalue__', jv)
        super(JObject, self).__init__()

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return object.__setattr__(self, name, value)

        if not hasattr(self, name):
            raise AttributeError("Field '%s' not found on Java '%s' object" %
                                 (name, self.__name__))

        try:
            attr = _jclass.typeLookup(type(self), name)
            if hasattr(attr, '__set__'):
                return attr.__set__(self, value)
        except AttributeError:
            pass
        raise AttributeError("Field '%s' is not settable on Java '%s' object" %
                             (name, self.__name__))

    def __str__(self):
        return self.__javavalue__.toString()

    def __unicode__(self):
        return self.__javavalue__.toUnicode()

    def __hash__(self):
        return self.hashCode()

    def __eq__(self, other):
        return self.equals(other)

    def __ne__(self, other):
        return not self.equals(other)


# Post load dependencies
_jclass._JObject = JObject
_jcustomizer._JObject = JObject


def _JObjectFactory(v=None, tp=None):
    """ Creates a Java object.

    If not specified type is determined based on the object.
    If type type is specified then then it tried to box it.
    """
    cls = None

    if isinstance(v, type):
        if hasattr(v, '__javaclass__'):
            cls = _jclass.JClass("java.lang.Class").__javaclass__
        else:
            raise TypeError("%s is not a java class." % v)

    # Automatically look up the type if not specified,
    if tp is None:
        return _jclass._getDefaultJavaObject(v)

    # If it is a string look up the class name,
    elif isinstance(tp, (str, _unicode)):
        return _jclass.JClass(tp)

    # Check if we are to box it,
    elif isinstance(tp, type):
        if hasattr(tp, '_java_boxed_class'):
            return tp._java_boxed_class
        elif hasattr(tp, '__javaclass__'):
            return _jclass.JClass(tp.__javaclass__)

    raise TypeError("Invalid type conversion to %s requested." % tp)


def defineJObjectFactory(name, jclass, proto, bases=(JObject,), members=None):
    """ Create a factory type such as JObject or JArray.

    Args:
        name (str): Name of the class to produce
        jclass (str): Name of the java class this should shadow.
        proto (type): Is a type from which the class methods will be based.
        bases (tuple): Bases for this meta class factory.
        members (dict): Any additional members for this class.

    """
    # Copy the members from the prototype
    if members == None:
        members = {}
    for p, v in proto.__dict__.items():
        if isinstance(v, (str, property)):
            members[p] = v
        elif callable(v):
            members[p] = v
        elif p == "__new__":
            members[p] = v

    res = None

    if jclass != None:
        members['__javaclass__'] = None

    # Create a new class
    res = _jclass.JClass(name, bases, members)

    if jclass != None:
        # Register this class to be initialized when jvm starts
        def jinit():
            type.__setattr__(res, '__javaclass__', _jpype.PyJPClass(jclass))
        _jinit.registerJVMInitializer(jinit)

    return res



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys

from . import _jproxy
from . import _jclass

__all__ = ['setupGuiEnvironment', 'shutdownGuiEnvironment']

# FIXME this is not documented
# FIXME this is darwin specific


def setupGuiEnvironment(cb):
    if _sys.platform == 'darwin':
        from PyObjCTools import AppHelper
        m = {'run': cb}
        proxy = _jproxy.JProxy('java.lang.Runnable', m)
        cbthread = _jclass.JClass("java.lang.Thread")(proxy)
        cbthread.start()
        AppHelper.runConsoleEventLoop()
    else:
        cb()


def shutdownGuiEnvironment():
    if _sys.platform == 'darwin':
        from PyObjCTools import AppHelper
        AppHelper.stopEventLoop()



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import jpype
import _jpype
from . import _jcustomizer
from . import _jobject

import sys as _sys

# Python2/3 support
if _sys.version_info > (3,):
    _long = int
else:
    _long = long

__all__ = []


@_jcustomizer.JImplementationFor("java.lang.Boolean", base=True)
class _JBoxedBoolean(int, _jobject.JObject):
    # Boolean is special because in python True and False are singletons,
    # thus it is not possible to make a wrapper properly act as a bool
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (bool, int, _long)):
            return int.__new__(cls, args[0])
        if hasattr(args[0], 'booleanValue'):
            return int.__new__(cls, args[0].booleanValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return int.__new__(cls, cls.booleanValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = int.__eq__
    __ne__ = int.__ne__

    if _sys.version_info < (3,):
        __cmp__ = int.__cmp__
    else:
        __lt__ = int.__lt__
        __gt__ = int.__gt__
        __le__ = int.__le__
        __ge__ = int.__ge__

    def __str__(self):
        if int(self) == 0:
            return str(False)
        return str(True)

    def __hash__(self):
        return int.__hash__(self)


@_jcustomizer.JImplementationFor("java.lang.Byte", base=True)
class _JBoxedByte(int, _jobject.JObject):
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (int, _long)):
            return int.__new__(cls, args[0])
        if hasattr(args[0], 'byteValue'):
            return int.__new__(cls, args[0].byteValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return int.__new__(cls, cls.byteValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = int.__eq__
    __ne__ = int.__ne__

    if _sys.version_info < (3,):
        __cmp__ = int.__cmp__
    else:
        __lt__ = int.__lt__
        __gt__ = int.__gt__
        __le__ = int.__le__
        __ge__ = int.__ge__

    def __hash__(self):
        return int.__hash__(self)


@_jcustomizer.JImplementationFor("java.lang.Short", base=True)
class _JBoxedShort(int, _jobject.JObject):
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (int, _long)):
            return int.__new__(cls, args[0])
        if hasattr(args[0], 'shortValue'):
            return int.__new__(cls, args[0].shortValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return int.__new__(cls, cls.shortValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = int.__eq__
    __ne__ = int.__ne__

    if _sys.version_info < (3,):
        __cmp__ = int.__cmp__
    else:
        __lt__ = int.__lt__
        __gt__ = int.__gt__
        __le__ = int.__le__
        __ge__ = int.__ge__

    def __hash__(self):
        return int.__hash__(self)


@_jcustomizer.JImplementationFor("java.lang.Integer", base=True)
class _JBoxedInteger(int, _jobject.JObject):
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (int, _long)):
            return int.__new__(cls, args[0])
        if hasattr(args[0], 'intValue'):
            return int.__new__(cls, args[0].intValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return int.__new__(cls, cls.intValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = int.__eq__
    __ne__ = int.__ne__

    if _sys.version_info < (3,):
        __cmp__ = int.__cmp__
    else:
        __lt__ = int.__lt__
        __gt__ = int.__gt__
        __le__ = int.__le__
        __ge__ = int.__ge__

    def __hash__(self):
        return int.__hash__(self)


@_jcustomizer.JImplementationFor("java.lang.Long", base=True)
class _JBoxedLong(_long, _jobject.JObject):
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (int, _long)):
            return _long.__new__(cls, args[0])
        if hasattr(args[0], 'longValue'):
            return _long.__new__(cls, args[0].longValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return _long.__new__(cls, cls.longValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = _long.__eq__
    __ne__ = _long.__ne__

    if _sys.version_info < (3,):
        __cmp__ = _long.__cmp__
    else:
        __lt__ = _long.__lt__
        __gt__ = _long.__gt__
        __le__ = _long.__le__
        __ge__ = _long.__ge__

    def __hash__(self):
        return _long.__hash__(self)


@_jcustomizer.JImplementationFor("java.lang.Float", base=True)
class _JBoxedFloat(float, _jobject.JObject):
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (int, _long, float)):
            return float.__new__(cls, args[0])
        if hasattr(args[0], 'longValue'):
            return float.__new__(cls, args[0].floatValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return float.__new__(cls, cls.floatValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = float.__eq__
    __ne__ = float.__ne__

    def __hash__(self):
        return float.__hash__(self)


@_jcustomizer.JImplementationFor("java.lang.Double", base=True)
class _JBoxedDouble(float, _jobject.JObject):
    def __new__(cls, *args):
        if len(args) != 1:
            raise TypeError("Invalid arguments")
        if isinstance(args[0], (int, _long, float)):
            return float.__new__(cls, args[0])
        if hasattr(args[0], 'longValue'):
            return float.__new__(cls, args[0].doubleValue())
        if isinstance(args[0], _jpype.PyJPValue):
            return float.__new__(cls, cls.doubleValue(args[0]))
        raise ValueError("Invalid arguments %s" % args[0])
    __eq__ = float.__eq__
    __ne__ = float.__ne__

    def __hash__(self):
        return float.__hash__(self)




#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import _jpype
import sys as _sys
from . import _jcustomizer
from . import _jclass
from . import _jobject

__all__ = ['JException']

if _sys.version > '3':
    _unicode = str
else:
    _unicode = unicode


class _JException(object):
    """ Base class for all ``java.lang.Throwable`` objects.

    When called as an object ``JException`` will produce a new exception class.  
    The arguments may either be a string or an existing Java throwable.  
    This functionality is deprecated as exception classes can be created with 
    ``JClass``.

    Use ``issubclass(cls, JException)`` to test if a class is derived 
    from ``java.lang.Throwable.``

    Use ``isinstance(obj, JException)`` to test if an object is a 
    ``java.lang.Throwable``.

    """
    def __new__(cls, *args, **kwargs):
        if cls == JException:
            import warnings
            if not hasattr(JException, '_warned'):
                warnings.warn("Using JException to construct an exception type is deprecated.",
                              category=DeprecationWarning, stacklevel=2)
                JException._warned = True
            return _JExceptionClassFactory(*args, **kwargs)
        return super(JException, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        if hasattr(self, '__javavalue__'):
            pass
        elif len(args) == 1 and isinstance(args[0], _jpype.PyJPValue):
            self.__javavalue__ = args[0]
        else:
            self.__javavalue__ = self.__class__.__javaclass__.newInstance(
                *args)
        super(Exception, self.__class__).__init__(self)

    def __str__(self):
        return str(self.toString())

    # Included for compatibility with JPype 0.6.3
    def message(self):
        return str(self.getMessage())

    # Included for compatibility with JPype 0.6.3
    def stacktrace(self):
        """ Get a string listing the stack frame.

        Returns:
          A string with the classic Java ``printStackTrace`` result.
        """
        StringWriter = _jclass.JClass("java.io.StringWriter")
        PrintWriter = _jclass.JClass("java.io.PrintWriter")
        sw = StringWriter()
        pw = PrintWriter(sw)
        self.printStackTrace(pw)
        pw.flush()
        r = sw.toString()
        sw.close()
        return r

    # For compatiblity with python exceptions
    @property
    def args(self):
        return self._jargs()
#    args = property(lambda self: self._jargs(), None)
#    """ Test doc string on property"""

    def _jargs(self):
        cause = self.getCause()
        if cause is None:
            return (str(self.getMessage()),)
        return (str(self.getMessage()), cause,)


JException = _jobject.defineJObjectFactory("JException", "java.lang.Throwable",
                                           _JException, bases=(Exception, _jobject.JObject))
_jcustomizer.registerClassBase('java.lang.Throwable', JException)


def _JExceptionClassFactory(tp):
    if isinstance(tp, (str, _unicode)):
        return _jclass.JClass(tp)
    if isinstance(tp, _jclass.JClass):
        return _jclass.JClass(tp.__javaclass__)
    raise TypeError(
        "JException requires a string or java throwable type, got %s." % tp)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import _jpype
import sys as _sys

_JP_BASES = {}
_JP_IMPLEMENTATIONS = {}

if _sys.version_info > (3,):
    _unicode = str
else:
    _unicode = unicode

__all__ = ['JImplementationFor']

# Forward declarations
_JObject = None
_JCLASSES = None


def registerClassBase(name, cls):
    """ (internal) Add an implementation for a class

    Use @JImplementationFor(cls, base=True) to access this.

    """
    if not issubclass(cls,  _JObject):
        raise TypeError("Classbases must derive from JObject")

    if name in _JP_BASES:
        _JP_BASES[name].append(cls)
    else:
        _JP_BASES[name] = [cls]

    # Changing the base class in python can break things,
    # so we will tag this as an error for now.
    if name in _JCLASSES:
        raise RuntimeError(
            "Base classes must be added before class is created")


def registerClassImplementation(classname, proto):
    """ (internal) Add an implementation for a class

    Use @JImplementationFor(cls) to access this.
    """
    if classname in _JP_IMPLEMENTATIONS:
        _JP_IMPLEMENTATIONS[classname].append(proto)
    else:
        _JP_IMPLEMENTATIONS[classname] = [proto]

    # If we have already created a class, apply it retroactively.
    if classname in _JCLASSES:
        _applyCustomizerPost(_JCLASSES[classname], proto)


def JImplementationFor(clsname, base=False):
    """ Decorator to define an implementation for a class.

    Applies to a class which will serve as a prototype as for the java class
    wrapper.  If it is registered as a base class, then the class must 
    derive from JObject.  Otherwise, the methods are copied from 
    the prototype to java class wrapper.

    The method ``__jclass_init__(cls)`` will be called with the constructed 
    class as the argument.  This call be used to set methods for all classes 
    that derive from the specified class.  Use ``type.__setattr__()`` to 
    alter the class methods.

    Using the prototype class as a base class is used mainly to support 
    classes which must be derived from a python type by design.  Use
    of a base class will produce a RuntimeError if the class has already
    been created.  

    For non-base class customizers, the customizer will be applied 
    retroactively if the class is already created.  Conflicts are
    resolved by the last customizer applied.

    Args:
      clsname (str): name of java class.
      base (bool, optional): if True this will be a base class. 
        Default is False.

    """
    if not isinstance(clsname, (str, _unicode)):
        raise TypeError("SuperFor requires a java classname string")

    def customizer(cls):
        if base:
            registerClassBase(clsname, cls)
        else:
            registerClassImplementation(clsname, cls)
        return cls
    return customizer


def _applyStickyMethods(cls, sticky):
    for method in sticky:
        attr = getattr(method, '__joverride__')
        rename = attr.get('rename', None)
        if rename:
            type.__setattr__(
                cls, rename, type.__getattribute__(cls, method.__name__))
        type.__setattr__(cls, method.__name__, method)


def _applyCustomizerImpl(members, proto, sticky, setter):
    """ (internal) Apply a customizer to a class.

    This "borrows" methods from a prototype class.
    Current behavior is:
     - Copy any string or property.
     - Copy any callable applying @JOverride
       if applicable with conflict renaming.
     - Copy __new__ method.

    """
    for p, v in proto.__dict__.items():
        if callable(v) or isinstance(v, (str, property, staticmethod, classmethod)):
            rename = "_"+p

            # Apply JOverride annotation
            attr = getattr(v, '__joverride__', None)
            if attr is not None:
                if attr.get('sticky', False):
                    sticky.append(v)
                rename = attr.get('rename', rename)

            # Apply rename
            if p in members and isinstance(members[p], (_jpype.PyJPField, _jpype.PyJPMethod)):
                setter(rename, members[p])

            setter(p, v)


def _applyAll(cls, method):
    applied = set()
    todo = [cls]
    while todo:
        c = todo.pop(0)
        if c in applied:
            continue
        todo.extend(c.__subclasses__())
        applied.add(c)
        method(c)


def _applyCustomizerPost(cls, proto):
    """ (internal) Customize a class after it has been created """
    sticky = []
    _applyCustomizerImpl(cls.__dict__, proto, sticky,
                         lambda p, v: type.__setattr__(cls, p, v))

    # Apply a customizer to all derived classes
    if '__jclass_init__' in proto.__dict__:
        method = proto.__dict__['__jclass_init__']
        _applyAll(cls, method)

    # Merge sticky into existing __jclass_init__
    if len(sticky) > 0:
        method = proto.__dict__.get('__jclass_init__', None)

        def init(cls):
            if method:
                method(cls)
            _applyStickyMethods(cls, sticky)
        type.__setattr__(cls, '__jclass_init__', init)
        _applyAll(cls, init)


def _applyCustomizers(name, jc, bases, members):
    """ (internal) Called by JClass and JArray to customize a newly created class."""
    # Apply base classes
    if name in _JP_BASES:
        for b in _JP_BASES[name]:
            bases.insert(0, b)

    # Apply implementations
    if name in _JP_IMPLEMENTATIONS:
        sticky = []
        for proto in _JP_IMPLEMENTATIONS[name]:
            _applyCustomizerImpl(members, proto, sticky,
                                 lambda p, v: members.__setitem__(p, v))

        if len(sticky) > 0:
            def init(cls):
                _applyStickyMethods(cls, sticky)
            members['__jclass_init__'] = init


def _applyInitializer(cls):
    """ (internal) Called after the class is created to apply any customizations
    required by inherited parents. 
    """
    if hasattr(cls, '__jclass_init__'):
        init = []
        for base in cls.__mro__:
            if '__jclass_init__' in base.__dict__:
                init.insert(0, base.__dict__['__jclass_init__'])
        for func in init:
            func(cls)



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************

import re
import _jpype
from . import _jclass

__all__ = ['JPackage']


class JPackage(object):
    """ Gateway for automatic importation of Java classes.

    This class allows structured access to Java packages and classes.
    This functionality has been replaced by ``jpype.imports``, but is still
    useful in some cases.

    Only the root of the package tree need be declared with the ``JPackage``
    constructor. Sub-packages will be created on demand.

    For example, to import the w3c DOM package:

    .. code-block:: python

      Document = JPackage('org').w3c.dom.Document

    Under some situations such as a missing jar the resulting object
    will be a JPackage object rather than the expected java class. This
    results in rather challanging debugging messages. Thus the
    ``jpype.imports`` module is preferred. To prevent these types of errors
    a package can be declares as ``strict`` which prevents expanding
    package names that do not comply with Java package name conventions.

    Args:
      path (str): Path into the Java class tree.
      strict (bool, optional): Requires Java paths to conform to the Java
        package naming convention. If a path does not conform and a class 
        with the required name is not found, the AttributeError is raise 
        to indicate that the class was not found.

    Example:

      .. code-block:: python

        # Alias into a library
        google = JPackage("com.google")

        # Access members in the library
        result = google.common.IntMath.pow(x,m)

    """

    def __init__(self, name, strict=False, pattern=None):
        self.__name = name
        self.__pattern = pattern
        if strict:
            self.__pattern = re.compile('[_a-z][_a-z0-9]')

    def __getattribute__(self, n):
        try:
            return object.__getattribute__(self, n)
        except AttributeError as ex:
            ex1 = ex
            pass

        if n.startswith("__"):
            raise ex1
        # not found ...

        # perhaps it is a class?
        subname = "{0}.{1}".format(self.__name, n)
        if not _jpype.isStarted():
            if n.startswith('_'):
                raise ex1
            import warnings
            warnings.warn(
                "JVM not started yet, can not inspect JPackage contents %s")
            return n

        # See if it is a Java class
        try:
            cc = _jclass.JClass(subname)
            self.__setattr__(n, cc, True)
            return cc
        except:
            pass

        # Check to see if this conforms to the required package name
        # convention, it not then we should not create a new package
        if self.__pattern and self.__pattern.match(n) == None:
            raise AttributeError(
                "Java package %s does not contain a class %s" % (self.__name, n))

        # Add package to the path
        cc = JPackage(subname, pattern=self.__pattern)
        self.__setattr__(n, cc, True)
        return cc

    def __setattr__(self, n, v, intern=False):
        if not n[:len('_JPackage')] == '_JPackage' \
           and not intern:  # NOTE this shadows name mangling
            raise RuntimeError("Cannot set attributes in a package {0}"
                               .format(n))
        object.__setattr__(self, n, v)

    def __str__(self):
        return "<Java package {0}>".format(self.__name)

    def __call__(self, *arg, **kwarg):
        raise TypeError("Package {0} is not Callable".format(self._name))



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys

import _jpype
from . import _jclass
from . import _jobject
from . import _jcustomizer

__all__ = ['JString']


class _JString(object):
    """ Base class for ``java.lang.String`` objects

    When called as a function, this class will produce a ``java.lang.String`` 
    object.  It can be used to test if an object is a Java string
    using ``isinstance(obj, JString)``.

    """
    def __new__(cls, *args, **kwargs):
        if cls == JString:
            cls = _jclass.JClass(JString.__javaclass__)
            return cls.__new__(cls, *args)
        return super(JString, cls).__new__(cls, *args, **kwargs)

    def __add__(self, other):
        return self.concat(other)

    def __eq__(self, other):
        if isinstance(other, JString):
            return self.equals(other)
        return str(self) == other

    def __ne__(self, other):
        if isinstance(other, JString):
            return not self.equals(other)
        return str(self) != other

    def __len__(self):
        return self.length()

    def __getitem__(self, i):
        return self.charAt(i)

    def __lt__(self, other):
        return self.compareTo(other) < 0

    def __le__(self, other):
        return self.compareTo(other) <= 0

    def __gt__(self, other):
        return self.compareTo(other) > 0

    def __ge__(self, other):
        return self.compareTo(other) >= 0

    def __contains__(self, other):
        return self.contains(other)

    def __hash__(self):
        return self.__str__().__hash__()

    def __repr__(self):
        return "'%s'"%self.__str__()


JString = _jobject.defineJObjectFactory(
    "JString", "java.lang.String", _JString)
_jcustomizer.registerClassBase('java.lang.String', JString)



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys
try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence


import _jpype
from . import _jclass
from . import _jobject
from . import _jstring
from . import _jcustomizer

if _sys.version > '3':
    unicode = str
    irange = range

__all__ = ['JArray']

_JARRAY_TYPENAME_MAP = {
    'boolean': 'Z',
    'byte': 'B',
    'char': 'C',
    'short': 'S',
    'int': 'I',
    'long': 'J',
    'float': 'F',
    'double': 'D',
}


class _JArray(object):
    """ Create a java array class for a Java type of a given dimension.

    This serves as a base type and factory for all Java array classes.
    The resulting Java array class can be used to construct a new
    array with a given size or members.

    JPype arrays support Python operators for iterating, length, equals, 
    not equals, subscripting, and limited slicing. They also support Java
    object methods, clone, and length property. Java arrays may not
    be resized, thus elements cannot be added nor deleted. Currently,
    applying the slice operator produces a new Python sequence.

    Example:
        .. code-block:: python

          # Define a new array class for ``int[]``
          IntArrayCls = JArray(JInt)

          # Create an array holding 10 elements 
          #   equivalent to Java ``int[] x=new int[10]``
          x = IntArrayCls(10)

          # Create a length 3 array initialized with [1,2,3]
          #   equivalent to Java ``int[] x = new int[]{1,2,3};``
          x = IntArrayCls([1,2,3])

          # Operate on an array
          print(len(x))
          print(x[0])
          print(x[:-2])
          x[1:]=(5,6)

          if isinstance(x, JArray):
               print("object is a java array")

          if issubclass(IntArrayCls, JArray):
               print("class is a java array type.")

    Args:
      javaClass (str,type): Is the type of element to hold in 
        the array.
      ndims (Optional,int): the number of dimensions of the array 
        (default=1)

    Returns:
      A new Python class that representing a Java array class.

    Raises:
      TypeError: if the component class is invalid or could not be found.

    Note: 
      javaClass can be specified in three ways:

        - as a string with the name of a java class.
        - as a Java primitive type such as ``jpype.JInt``.
        - as a Java class type such as ``java.lang.String``.


    """
    def __new__(cls, *args, **kwargs):
        if cls == JArray:
            return _JArrayNewClass(*args, **kwargs)
        return super(JArray, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        if hasattr(self, '__javavalue__'):
            self.__javaarray__ = _jpype.PyJPArray(self.__javavalue__)
            return

        if len(args) != 1:
            raise TypeError(
                "Array classes only take 2 parameters, {0} given"
                .format(len(args) + 1))

        if isinstance(args[0], _jpype.PyJPValue):
            self.__javavalue__ = args[0]
            self.__javaarray__ = _jpype.PyJPArray(self.__javavalue__)
            return

        values = None
        if _isIterable(args[0]):
            sz = len(args[0])
            values = args[0]
        else:
            sz = args[0]

        self.__javavalue__ = self.__class__.__javaclass__.newInstance(sz)
        self.__javaarray__ = _jpype.PyJPArray(self.__javavalue__)

        if values is not None:
            self.__javaarray__.setArraySlice(0, sz, values)

    def __str__(self):
        return str(tuple(self))

    @property
    def length(self):
        """ Get the length of a Java array

        This method is provided for compatiblity with Java syntax.
        Generally, the Python style ``len(array)`` should be preferred.
        """
        return self.__len__()

    def __len__(self):
        return self.__javaarray__.getArrayLength()

    def __iter__(self):
        return _JavaArrayIter(self)

    def __getitem__(self, ndx):
        if isinstance(ndx, slice):
            start, stop, step = ndx.indices(len(self))
            if step != 1:
                raise NotImplementedError("Slicing with step unimplemented")
            return self.__getslice__(start, stop)
        return self.__javaarray__.getArrayItem(ndx)

    def __setitem__(self, ndx, val):
        if isinstance(ndx, slice):
            start, stop, step = ndx.indices(len(self))
            if step != 1:
                # Iterate in python if we need to step
                indices = irange(start, stop, step)
                for index, value in zip(indices, val):
                    self[index] = value
            else:
                self.__setslice__(start, stop, val)
            return
        self.__javaarray__.setArrayItem(ndx, val)

    def __getslice__(self, i, j):
        if j == _sys.maxsize:
            j = self.__javaarray__.getArrayLength()
        return self.__javaarray__.getArraySlice(i, j)

    def __setslice__(self, i, j, v):
        if j == _sys.maxsize:
            j = self.__javaarray__.getArrayLength()
        self.__javaarray__.setArraySlice(i, j, v)

    def __eq__(self, other):
        if hasattr(other, '__javavalue__'):
            return self.equals(other)
        try:
            return self.equals(self.__class__(other))
        except TypeError:
            return False

    def __ne__(self, other):
        if hasattr(other, '__javavalue__'):
            return not self.equals(other)
        try:
            return self.equals(self.__class__(other))
        except TypeError:
            return True

    def __hash__(self):
        return self.hashCode()

    def clone(self):
        """ Clone the Java array.

        Create a "shallow" copy of a Java array. For a
        single dimensional array of primitives, the cloned array is
        complete independent copy of the original. For objects or 
        multidimensional arrays, the new array is a copy which points
        to the same members as the original.

        To obtain a deep copy of a Java array, use Java serialize and
        deserialize operations to duplicate the entire array and 
        contents. In order to deep copy, the objects must be 
        Serializable.

        Returns:
            A shallow copy of the array.
        """
        return _jclass.JClass("java.util.Arrays").copyOf(self, len(self))



JArray = _jobject.defineJObjectFactory("JArray", None, _JArray)


def _JArrayNewClass(cls, ndims=1):
    """ Convert a array class description into a JArray class."""
    jc = _jclass._toJavaClass(cls)

    if jc.isPrimitive():
        # primitives need special handling
        typename = ('['*ndims)+_JARRAY_TYPENAME_MAP[jc.getCanonicalName()]
    elif jc.isArray():
        typename = ('['*ndims)+str(_jobject.JObject(jc).getName())
    else:
        typename = ('['*ndims)+'L'+str(_jobject.JObject(jc).getName())+';'

    return _jclass.JClass(typename)


# FIXME JavaArrayClass likely should be exposed for isinstance, issubtype
# FIXME are these not sequences?  They act like sequences but are they
# connected to collections.Sequence
# has: __len__, __iter__, __getitem__
# missing: __contains__ (required for in)
# Cannot be Mutable because java arrays are fixed in length

def _isIterable(obj):
    if isinstance(obj, Sequence):
        return True
    if hasattr(obj, '__len__') and hasattr(obj, '__iter__'):
        return True
    return False


class _JavaArrayIter(object):
    def __init__(self, a):
        self._array = a
        self._ndx = -1

    def __iter__(self):
        return self

    def __next__(self):
        self._ndx += 1
        if self._ndx >= len(self._array):
            raise StopIteration
        return self._array[self._ndx]

    next = __next__

# **********************************************************
# Char array customizer


@_jcustomizer.JImplementationFor("byte[]")
@_jcustomizer.JImplementationFor("char[]")
class _JCharArray(object):
    def __str__(self):
        return str(_jstring.JString(self))

    def __unicode__(self):
        return unicode(_jstring.JString(self))

    def __eq__(self, other):
        if hasattr(other, '__javavalue__'):
            return self.equals(other)
        elif isinstance(other, (str, unicode)):
            return self[:] == other
        try:
            return self.equals(self.__class__(other))
        except TypeError:
            return False



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
from . import _jcustomizer
import sys as _sys


@_jcustomizer.JImplementationFor('java.lang.Comparable')
class _JComparable(object):
    if _sys.version_info < (3,):
        def __cmp__(self, o):
            return self.compareTo(o)

    def __eq__(self, o):
        return self.compareTo(o) == 0

    def __ne__(self, o):
        return self.compareTo(o) != 0

    def __gt__(self, o):
        return self.compareTo(o) > 0

    def __lt__(self, o):
        return self.compareTo(o) < 0

    def __ge__(self, o):
        return self.compareTo(o) >= 0

    def __le__(self, o):
        return self.compareTo(o) <= 0



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys
import _jpype
from ._pykeywords import pysafe
from . import _jcustomizer

__all__ = ['JClass', 'JInterface', 'JOverride']

if _sys.version > '3':
    _unicode = str
    _long = int
else:
    _unicode = unicode
    _long = long

_JObject = None

_java_lang_RuntimeException = None
_java_ClassLoader = None
_java_lang_Class = None
_java_lang_Object = None

_JCLASSES = {}
_JP_TYPE_CLASSES = {}
_JP_OBJECT_CLASSES = {}

_jcustomizer._JCLASSES = _JCLASSES


def _initialize():
    global _java_ClassLoader

    # Due to bootstrapping, Object and Class must be defined first.
    global _java_lang_Object, _java_lang_Class
    _java_lang_Object = JClass("java.lang.Object")
    _java_lang_Class = JClass("java.lang.Class")

    _java_ClassLoader = JClass('java.lang.ClassLoader').getSystemClassLoader()
    if not _java_ClassLoader:
        raise RuntimeError("Unable to load Java SystemClassLoader")

    global _java_lang_RuntimeException
    _java_lang_RuntimeException = JClass("java.lang.RuntimeException")

    # Preload needed classes
    java_lang_Boolean = JClass("java.lang.Boolean")
    java_lang_Long = JClass("java.lang.Long")
    java_lang_Double = JClass("java.lang.Double")
    java_lang_String = JClass("java.lang.String")

    global _JP_OBJECT_CLASSES
    _JP_OBJECT_CLASSES[bool] = java_lang_Boolean
    _JP_OBJECT_CLASSES[int] = java_lang_Long
    _JP_OBJECT_CLASSES[_long] = java_lang_Long
    _JP_OBJECT_CLASSES[float] = java_lang_Double
    _JP_OBJECT_CLASSES[str] = java_lang_String
    _JP_OBJECT_CLASSES[_unicode] = java_lang_String
    _JP_OBJECT_CLASSES[type] = _java_lang_Class
    _JP_OBJECT_CLASSES[_jpype.PyJPClass] = _java_lang_Class
    _JP_OBJECT_CLASSES[object] = _java_lang_Object


def JOverride(*args, **kwargs):
    """Annotation to denote a method as overriding a Java method.

    This annotation applies to customizers, proxies, and extension
    to Java class. Apply it to methods to mark them as implementing
    or overriding Java methods.  Keyword arguments are passed to the
    corresponding implementation factory.

    Args:
      sticky=bool: Applies a customizer method to all derived classes.

    """
    # Check if called bare
    if len(args) == 1 and callable(args[0]):
        object.__setattr__(args[0], "__joverride__", {})
        return args[0]
     # Otherwise apply arguments as needed

    def modifier(method):
        object.__setattr__(method, "__joverride__", kwargs)
        return method
    return modifier


class JClass(type):
    """Meta class for all java class instances.

    JClass when called as an object will contruct a new java Class wrapper.

    All python wrappers for java classes derived from this type.
    To test if a python class is a java wrapper use
    ``isinstance(obj, jpype.JClass)``.

    Args:
      className (str): name of a java type.

    Keyword Args:
      loader (java.lang.ClassLoader): specifies a class loader to use
        when creating a class.
      initialize (bool): Passed to class loader when loading a class
        using the class loader.

    Returns:
      JavaClass: a new wrapper for a Java class

    Raises:
      TypeError: if the component class is invalid or could not be found.
    """
    @property
    def class_(self):
        return _JObject(self.__javaclass__)

    def __new__(cls, *args, **kwargs):
        if len(args) == 1:
            return _JClassNew(args[0], **kwargs)
        return super(JClass, cls).__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            return
        super(JClass, self.__class__).__init__(self, *args)

    def __getattribute__(self, name):
        if name.startswith('_'):
            if name == "__doc__":
                return _jclassDoc(self)
            return type.__getattribute__(self, name)
        attr = type.__getattribute__(self, name)
        if isinstance(attr, _jpype.PyJPMethod):
            return attr
        if isinstance(attr, property):
            raise AttributeError("Field is not static")
        return attr

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return type.__setattr__(self, name, value)

        if not hasattr(self, name):
            raise AttributeError("Static field '%s' not found on Java '%s' class" %
                                 (name, self.__name__))

        try:
            attr = typeLookup(self, name)
            if hasattr(attr, '__set__'):
                return attr.__set__(self, value)
        except AttributeError:
            pass
        raise AttributeError("Static field '%s' is not settable on Java '%s' class" %
                             (name, self.__name__))

    def mro(cls):
        # Bases is ordered by (user, extend, interfaces)
        # Thus we maintain order of parents whereever possible
        # Breadth first inclusion of the parents
        parents = list(cls.__bases__)
        out = [cls]

        # Until we run out of parents
        while parents:

            # Test if the first is a leaf of the tree
            front = parents.pop(0)
            for p in parents:
                if issubclass(p, front):
                    parents.append(front)
                    front = None
                    break

            if not front:
                # It is not a leaf, we select another candidate
                continue

            # It is a leaf, so add it to the parents
            out.append(front)

            # Place the immediate parents of newest in the head of the list
            prev = parents
            parents = list(front.__bases__)

            # Include the remaining that we still need to consider
            parents.extend([b for b in prev if not b in parents])

        # JavaObjects are not interfaces, so we need to remove the JavaInterface inheritance
        if _JObject in out and JInterface in out:
            out.remove(JInterface)

        return out

    def __repr__(self):
        return "<java class '%s'>"%(self.__name__)


def _JClassNew(arg, loader=None, initialize=True):
    params = None
    if isinstance(arg, str):
        if arg.endswith(">"):
            i = arg.find("<")
            params = arg[i+1:-1]
            arg = arg[:i]
            if params:
                params = params.split(',')
            else:
                params = []
            # FIXME we can check the geneic parameters better.

    if loader and isinstance(arg, str):
        arg = _java_lang_Class.forName(arg, initialize, loader)

    if isinstance(arg, _jpype.PyJPClass):
        javaClass = arg
    else:
        javaClass = _jpype.PyJPClass(arg)

    if javaClass is None:
        raise _java_lang_RuntimeException("Java class '%s' not found" % name)

    # Lookup the class name
    name = javaClass.getCanonicalName()

    # See if we have an existing class in the cache
    if name in _JCLASSES:
        cls = _JCLASSES[name]
    else:
        cls = _JClassFactory(name, javaClass)

    if params is not None:
        acceptParams = len(cls.class_.getTypeParameters())
        if acceptParams == 0:
            raise TypeError(
                "Java class '%s' does not take parameters" % (cls.__name__))
        if len(params) > 0 and len(params) != len(cls.class_.getTypeParameters()):
            raise TypeError(
                "Java generic class '%s' length mismatch" % (cls.__name__))
    return cls


class JInterface(object):
    """Base class for all Java Interfaces.

    ``JInterface`` is serves as the base class for any java class that is
    a pure interface without implementation. It is not possible to create
    a instance of a java interface. The ``mro`` is hacked such that
    ``JInterface`` does not appear in the tree of objects implement an
    interface.

    Example:

    .. code-block:: python

       if issubclass(java.util.function.Function, jpype.JInterface):
          print("is interface")


        Use ``isinstance(obj, jpype.JavaInterface)`` to test for a interface.
    """
    @property
    def class_(self):
        return _JObject(self.__javaclass__)

    def __new__(cls, *args, **kwargs):
        return super(JInterface, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], _jpype.PyJPValue):
            object.__setattr__(self, '__javavalue__', args[0])
        elif not hasattr(self, '__javavalue__'):
            raise JClass("java.lang.InstantiationException")(
                "`%s` is an interface." % str(self.class_.getName()))
        super(JInterface, self).__init__()

    def __str__(self):
        return self.toString()

    def __hash__(self):
        return self.hashCode()

    def __eq__(self):
        return self.equals(o)

    def __ne__(self):
        return not self.equals(o)


def _JClassFactory(name, jc):
    from . import _jarray

    # Set up bases
    bases = []
    bjc = jc.getSuperClass()
    if name == 'java.lang.Object':
        bases.append(_JObject)
    elif jc.isArray():
        bases.append(_jarray.JArray)
    elif jc.isPrimitive():
        bases.append(object)
    elif bjc is not None:
        bases.append(JClass(bjc))
    elif bjc is None:
        bases.append(JInterface)
    itf = jc.getInterfaces()
    for ic in itf:
        bases.append(JClass(ic))

    # Set up members
    members = {
        "__javaclass__": jc,
        "__name__": name,
    }
    fields = jc.getClassFields()
    for i in fields:
        fname = pysafe(i.getName())
        members[fname] = i
    for jm in jc.getClassMethods():
        members[pysafe(jm.__name__)] = jm

    # Apply customizers
    _jcustomizer._applyCustomizers(name, jc, bases, members)
    res = JClass(name, tuple(bases), members)
    _JCLASSES[name] = res

    # Post customizers
    _jcustomizer._applyInitializer(res)

    # Attach public inner classes we find
    #   Due to bootstrapping, we must wait until java.lang.Class is defined
    #   before we can access the class structures.
    if _java_lang_Class:
        for cls in res.class_.getDeclaredClasses():
            if cls.getModifiers() & 1 == 0:
                continue
            cls2 = JClass(cls)
            type.__setattr__(res, str(cls.getSimpleName()), cls2)

    return res

# **********************************************************


def _toJavaClass(tp):
    """(internal) Converts a class type in python into a internal java class.

    Used mainly to support JArray.

    The type argument will operate on:
     - (str) lookup by class name or fail if not found.
     - (JClass) just returns the java type.
     - (type) uses a lookup table to find the class.
    """
    # if it a string
    if isinstance(tp, (str, _unicode)):
        return JClass(tp).__javaclass__

    if isinstance(tp, _jpype.PyJPClass):
        return tp

    if not isinstance(tp, type):
        raise TypeError(
            "Argument must be a class, java class, or string representing a java class")

    # See if it a class type
    try:
        return tp.__javaclass__
    except AttributeError:
        pass

    try:
        return _JP_TYPE_CLASSES[tp].__javaclass__
    except KeyError:
        pass

    raise TypeError("Unable to find class for '%s'" % tp.__name__)


def _getDefaultJavaObject(obj):
    global _JP_OBJECT_CLASSES
    try:
        return _JP_OBJECT_CLASSES[type(obj)]
    except KeyError:
        pass

    try:
        return obj._java_boxed_class
    except AttributeError:
        pass

    if isinstance(obj, _jpype.PyJPClass):
        return _java_lang_Class

    # We need to check this first as values have both
    # __javavalue__ and __javaclass__
    if hasattr(obj, '__javavalue__'):
        return JClass(obj.__javaclass__)

    if hasattr(obj, '__javaclass__'):
        return _java_lang_Class

    if hasattr(obj, '__javaproxy__'):
        return _java_lang_Object

    if obj == None:
        return _java_lang_Object

    raise TypeError(
        "Unable to determine the default type of `{0}`".format(obj.__class__))


def typeLookup(tp, name):
    """Fetch a descriptor from the inheritance tree.

    This uses a cache to avoid additional cost when accessing items deep in
    the tree multiple times.
    """
    # TODO this cache may have interactions with retroactive
    # customizers. We should likely clear the cache if
    # the class is customized.
    try:
        cache = tp.__dict__['_cache']
    except:
        cache = {}
        type.__setattr__(tp, '_cache', cache)
    if name in cache:
        return cache[name]

    # Drill down the tree searching all parents for a field
    for cls in tp.__mro__:
        if name in cls.__dict__:
            obj = cls.__dict__[name]
            cache[name] = obj
            return obj

    cache[name] = None
    return None


def _jclassDoc(cls):
    """Generator for JClass.__doc__ property

    Parameters:
       cls (JClass): class to document.

    Returns:
      The doc string for the class.
    """
    from textwrap import TextWrapper
    jclass = cls.class_
    out = []
    out.append("Java class '%s'" % (jclass.getName()))
    out.append("")

    sup = jclass.getSuperclass()
    if sup:
        out.append("  Extends:")
        out.append("    %s" % sup.getName())
        out.append("")

    intfs = jclass.getInterfaces()
    if intfs:
        out.append("  Interfaces:")
        words = ", ".join([str(i.getCanonicalName()) for i in intfs])
        wrapper = TextWrapper(initial_indent='        ',
                              subsequent_indent='        ')
        out.extend(wrapper.wrap(words))
        out.append("")

    ctors = jclass.getDeclaredConstructors()
    if ctors:
        exceptions = []
        name = jclass.getSimpleName()
        ctordecl = []
        for ctor in ctors:
            modifiers = ctor.getModifiers()
            if not modifiers & 1:
                continue
            params = ", ".join([str(i.getCanonicalName())
                                for i in ctor.getParameterTypes()])
            ctordecl.append("    * %s(%s)" % (name, params))
            exceptions.extend(ctor.getExceptionTypes())
        if ctordecl:
            out.append("  Constructors:")
            out.extend(ctordecl)
            out.append("")
        if exceptions:
            out.append("  Raises:")
            for exc in set(exceptions):
                out.append("    %s: from java" % exc.getCanonicalName())
            out.append("")

    fields = jclass.getDeclaredFields()
    if fields:
        fielddesc = []
        for field in fields:
            modifiers = field.getModifiers()
            if not modifiers & 1:
                continue
            fieldInfo = []
            if modifiers & 16:
                fieldInfo.append("final")
            if modifiers & 8:
                fieldInfo.append("static")
            if field.isEnumConstant():
                fieldInfo.append("enum constant")
            else:
                fieldInfo.append("field")
            fielddesc.append("    %s (%s): %s" % (field.getName(),
                                                  field.getType().getName(),
                                                  " ".join(fieldInfo)))
        if fielddesc:
            out.append("  Attributes:")
            out.extend(fielddesc)
            out.append("")

    return "\n".join(out)


def _jmethodDoc(method, cls, overloads):
    """Generator for PyJPMethod.__doc__ property

    Parameters:
      method (PyJPMethod): method to generate doc string for.
      cls (java.lang.Class): Class holding this method dispatch.
      overloads (java.lang.reflect.Method[]): tuple holding all the methods
        that are served by this method dispatch.

    Returns:
      The doc string for the method dispatch.
    """
    from textwrap import TextWrapper
    out = []
    out.append("Java method dispatch '%s' for '%s'" %
               (method.__name__, cls.getName()))
    out.append("")
    exceptions = []
    returns = []
    methods = []
    classmethods = []
    for ov in overloads:
        modifiers = ov.getModifiers()
        exceptions.extend(ov.getExceptionTypes())
        returnName = ov.getReturnType().getCanonicalName()
        params = ", ".join([str(i.getCanonicalName())
                            for i in ov.getParameterTypes()])
        if returnName != "void":
            returns.append(returnName)
        if modifiers & 8:
            classmethods.append("    * %s %s(%s)" %
                                (returnName, ov.getName(), params))
        else:
            methods.append("    * %s %s(%s)" %
                           (returnName, ov.getName(), params))
    if classmethods:
        out.append("  Static Methods:")
        out.extend(classmethods)
        out.append("")

    if methods:
        out.append("  Virtual Methods:")
        out.extend(methods)
        out.append("")

    if exceptions:
        out.append("  Raises:")
        for exc in set(exceptions):
            out.append("    %s: from java" % exc.getCanonicalName())
        out.append("")

    if returns:
        out.append("  Returns:")
        words = ", ".join([str(i) for i in set(returns)])
        wrapper = TextWrapper(initial_indent='    ',
                              subsequent_indent='    ')
        out.extend(wrapper.wrap(words))
        out.append("")

    return "\n".join(out)


def _jmethodAnnotation(method, cls, overloads):
    """Generator for ``PyJPMethod.__annotation__`` property

    Parameters:
      method (PyJPMethod): method to generate annotations for.
      cls (java.lang.Class): Class holding this method dispatch.
      overloads (java.lang.reflect.Method[]): tuple holding all the methods
        that are served by this method dispatch.

    Returns:
      The dict to use for type annotations.
    """
    returns = []

    # Special handling if we have 1 overload
    if len(overloads)==1:
        ov = overloads[0]
        out = {}
        for i,p in enumerate(ov.getParameterTypes()):
            out['arg%d'%i] = JClass(p)
        out['return'] = JClass(ov.getReturnType())
        return out

    # Otherwise, we only get the return
    for ov in overloads:
        returns.append(ov.getReturnType())
    returns = set(returns)
    if len(returns) == 1:
        return {"return": JClass([i for i in returns][0])}
    return {}

def _jmethodCode(method):
    def call(*args):
       return method.__call__(*args)
    return call

_jpype.setResource('GetClassMethod', _JClassNew)
_jpype.setResource('GetMethodDoc', _jmethodDoc)
_jpype.setResource('GetMethodAnnotations', _jmethodAnnotation)
_jpype.setResource('GetMethodCode', _jmethodCode)



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
""" 
JPype Beans Module
------------------

This customizer finds all occurances of methods with get or set and converts
them into Python properties. This behavior is sometimes useful in programming
with JPype with interactive shells, but also leads to a lot of confusion.
Is this class exposing a variable or is this a property added JPype.  It was
the default behavior until 0.7. 

As an unnecessary behavior that violates both the Python principle 
*"There should be one-- and preferably only one --obvious way to do it."* and
the C++ principle *"You only pay for what you use"*. Thus this misfeature
was removed from the distribution as a default. However, given that it is
at times useful to have methods appear as properties, it was moved to a
an optional module.

To use beans as properties:

.. code-block:: python

  import jpype.beans

The beans property modification is a global behavior and applies retroactively
to all classes currently loaded.  Once started it can never be undone.
"""
import _jpype
from . import _jcustomizer
from ._pykeywords import pysafe


def _extract_accessor_pairs(members):
    """Extract pairs of corresponding property access methods
    (getter and setter) from a Java class's members (attributes).

    If a public method with a property's name exists no pair for
    that property will be extracted.

    Returns a dictionary with the property name as key and a tuple
    of (getter method, setter method) as value. A tuple element
    value might be `None` if only a getter or only a setter
    exists.
    """
    accessor_pairs = {}

    for name, member in members.items():
        if not isinstance(member, _jpype.PyJPMethod) or len(name) <= 3:
            continue
        if name == "getClass":
            continue

        if name.startswith("get") and member.isBeanAccessor():
            property_name = name[3].lower() + name[4:]
            try:
                pair = accessor_pairs[property_name]
                pair[0] = member
            except KeyError:
                accessor_pairs[property_name] = [member, None]
        elif name.startswith("set") and member.isBeanMutator():
            property_name = name[3].lower() + name[4:]
            try:
                pair = accessor_pairs[property_name]
                pair[1] = member
            except KeyError:
                accessor_pairs[property_name] = [None, member]
    return accessor_pairs


@_jcustomizer.JImplementationFor("java.lang.Object")
class _BeansCustomizer(object):
    """ Add properties for get/set Bean patterns found in classes.  """
    def __jclass_init__(cls):
        accessor_pairs = _extract_accessor_pairs(cls.__dict__)
        for attr_name, (getter, setter) in accessor_pairs.items():
            attr_name = pysafe(attr_name)

            # Don't mess with an existing member
            if attr_name in cls.__dict__:
                continue

            # Add the property
            type.__setattr__(cls, attr_name, property(getter, setter))



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import _jpype
from . import _core

# FIXME this class is entirely replaced with calls to .class_ and should likely be deprecated.


@_core.deprecated("class_.getConstructors()")
def getConstructors(clas):
    return clas.class_.getConstructors()[:]


@_core.deprecated("class_.getDeclaredConstructors()")
def getDeclaredConstructors(clas):
    return clas.class_.getDeclaredConstructors()[:]


@_core.deprecated("class_.getDeclaredFields()")
def getDeclaredFields(clas):
    '''Returns an array of Field objects reflecting all the fields declared by the class or interface represented by this Class object.'''
    return clas.class_.getDeclaredFields()[:]


@_core.deprecated("class_.getDeclaredMethods()")
def getDeclaredMethods(clas):
    '''Returns an array of Method objects reflecting all the methods declared by the class or interface represented by this Class object.'''
    return clas.class_.getDeclaredMethods()[:]


@_core.deprecated("class_.getFields()")
def getFields(clas):
    '''Returns an array containing Field objects reflecting all the accessible public fields of the class or interface represented by this Class object.'''
    return clas.class_.getFields()[:]


@_core.deprecated("class_.getMethods()")
def getMethods(clas):
    '''Returns an array containing Method objects reflecting all the public member methods of the class or interface represented by this Class object, including those declared by the class or interface and those inherited from superclasses and superinterfaces.'''
    return clas.class_.getMethods()[:]


@_core.deprecated("class_.getModifiers()")
def getModifiers(clas):
    '''Returns the Java language modifiers for this class or interface, encoded in an integer.'''
    return clas.class_.getModifiers()



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import sys as _sys
import _jpype

__all__ = ['convertToDirectBuffer']


if _sys.version_info < (2, 7):
    _mem_view = _jpype.memoryview  # own memoryview implementation
else:
    _mem_view = memoryview

# FIXME this is not documented.
# FIXME there are no tests of this method


def convertToDirectBuffer(obj):
    __doc__ = '''Efficiently convert all array.array and numpy ndarray types, string and unicode to java.nio.Buffer objects.'''

    memoryview_of_obj = _mem_view(obj)

    if memoryview_of_obj.readonly:
        raise ValueError(
            "Memoryview must be writable for wrapping in a byte buffer")

    return _jpype.convertToDirectBuffer(memoryview_of_obj)



# *****************************************************************************
#   Copyright 2004-2008 Steve Menard
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# *****************************************************************************
import _jpype
from ._jpackage import *
from ._jproxy import *
from ._core import *
from ._gui import *
from ._classpath import *
from ._jclass import *
from .types import *
from ._jcustomizer import *
from . import reflect
from . import nio
from . import types
from ._jcustomizer import *


__all__ = ['java', 'javax', 'JException', 'JOverride']
__all__.extend(_core.__all__)
__all__.extend(_classpath.__all__)
__all__.extend(types.__all__)
__all__.extend(_jproxy.__all__)
__all__.extend(_jpackage.__all__)
__all__.extend(_jclass.__all__)
__all__.extend(_jcustomizer.__all__)
__all__.extend(_gui.__all__)

__version_info__ = (0, 7, 1)
__version__ = ".".join(str(i) for i in __version_info__)


@_core.deprecated
def JIterator(it):
    """Deprecated"""
    return it


# FIXME these should be deprecated. The old JPackage system is only for
#  python2 series and generates lots of deceptive classes.  At some point
#  these two are going to have to go away.
java = JPackage("java", strict=True)
javax = JPackage("javax", strict=True)

JMethod = _jpype.PyJPMethod
JField = _jpype.PyJPField




import os as _os
import sys as _sys
import glob as _glob

__all__ = ['addClassPath', 'getClassPath']

_CLASSPATHS = []
_SEP = _os.path.pathsep
if _sys.platform == 'cygwin':
    _SEP = ';'


# Cygwin needs to convert to windows paths
if _sys.platform == 'cygwin':
    _root = None

    def _get_root():
        global _root
        if _root != None:
            return _root
        import subprocess
        proc = subprocess.Popen("cygpath -wa /", shell=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, close_fds=True)
        _root = proc.stdout.read().strip().decode('utf-8')
        return _root

    def _splitpath(path):
        parts = []
        (path, tail) = _os.path.split(path)
        while path and tail:
            parts.insert(0, tail)
            (path, tail) = _os.path.split(path)
        return parts

    def _posix2win(directory):
        root = _get_root()
        directory = _os.path.abspath(directory)
        paths = _splitpath(directory)
        if paths[0] == "cygdrive":
            paths.pop(0)
            drive = paths.pop(0)
            paths.insert(0, "%s:" % drive)
            return '\\'.join(paths)
        paths.insert(0, root)
        return '\\'.join(paths)
    # needed for testing
    __all__.append("_posix2win")


def addClassPath(path1):
    """ Add a path to the Java class path

    Classpath items can be a java, a directory, or a
    glob pattern.

    Arguments:
      path(str): 

    """
    global _CLASSPATHS

    path1 = _os.path.abspath(path1)
    if _sys.platform == 'cygwin':
        path1 = _posix2win(path1)
    _CLASSPATHS.append(str(path1))


def getClassPath(env=True):
    """ Get the full java class path.

    Includes user added paths and the environment CLASSPATH.

    Arguments:
      env(Optional, bool): If true then environment is included. 
        (default True)
    """
    global _CLASSPATHS
    global _SEP

    # Merge the evironment path
    classPath = list(_CLASSPATHS)
    envPath = _os.environ.get("CLASSPATH")
    if env and envPath:
        classPath.extend(envPath.split(_SEP))

    out = []
    for path in classPath:
        if path == '':
            continue
        if path.endswith('*'):
            paths = _glob.glob(path+".jar")
            if len(path) == 0:
                continue
            out.extend(paths)
        else:
            out.append(path)
    return _SEP.join(out)
